# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from .ndvar import Butterworth, concatenate, convolve, correlation_coefficient, cross_correlation, cwt_morlet, dss, erode, filter_data, find_intervals, find_peaks, frequency_response, gaussian, label_operator, maximum, minimum, neighbor_correlation, normalize_in_cells, powerlaw_noise, psd_welch, rename_dim, resample, segment, set_adjacency, set_parc, set_time, set_tmin
